/*
Copyright (C) 2006-2008 Keio University
(Kris Popendorf) <comp@bio.keio.ac.jp> (2006)

This file is part of Murasaki.

Murasaki is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Murasaki is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Murasaki.  If not, see <http://www.gnu.org/licenses/>.
*/

////////
// super-memory-miserly hash list storage
// Kris Popendorf
///////

#ifndef ECOHASH__H
#define ECOHASH__H

#include "ecolist.h"
#include "sequence.h"

using namespace std;

class EcoHash : public Hash {
 public:
  typedef Ecolist::val_type val_type;

  //abstract parts
  void clear();
  void add(const HashKey key,const HashVal &val);
  void lookup(HashKey key,const HashVal &a,LocList &locList);
  void lookup(HashKey key,LocList &locList);//debug only
  bool emptyAt(const HashKey key);
  word sizeAt(const HashKey key);
 
  //for computing memory costs
  static const word linear_cost(word c) {
    return ((word)(Ecolist::seqbits+Ecolist::idxbits))*(c/8);
  }
  static const word bucket_prep_cost(word c)
  {
    return (Ecolist::minBlocks*sizeof(word)+sizeof(Ecolist))*c+
      c/WORDSIZE*sizeof(word);
  }

  // dump / load may be overloaded
  //  void dump(ostream &os);
  //  void load(istream &is);

  //the boring junk
  EcoHash(BitSequence *pat);
  ~EcoHash();

  //muchos boring
  inline static Location val2loc(const val_type& a){return Location(a.first,a.second);}
  inline static val_type loc2val(const Location& a){return val_type(a.seqno,a.pos);}
  static bool lessthan(const val_type& a,const val_type& b);

 protected:
  Ecolist *fasta;

  class bitmap {
  public:
    inline bitmap(word c):
    map(new word[(c-1)/WORDSIZE+1]),
      size((c-1)/WORDSIZE+1)
	{
	  memset(map,0,sizeof(word)*size);
	}
    
    inline void clear(){
      memset(map,0,sizeof(word)*size);
    }

    inline bool checkAndSet(word hash){
      word idx=hash/WORDSIZE;
      word mask=((word)1U)<<(MODWORDSIZE(hash));
      bool res=(map[idx] & mask);
      if(!res){ //it seems that leaving this check in is actually faster!
	map[idx]|=mask;
      }
      return res;
    }
    inline bool check(word hash) const{
      word idx=hash/WORDSIZE;
      word mask=((word)1U)<<(MODWORDSIZE(hash));
      bool res=(map[idx] & mask);
      return res;
    }
    inline void set(word hash){
      word idx=hash/WORDSIZE;
      word mask=((word)1U)<<(MODWORDSIZE(hash));
      if(map[idx] & mask)
	map[idx]|=mask;
    }
    inline void flip(word hash){
      word idx=hash/WORDSIZE;
      word mask=((word)1U)<<(MODWORDSIZE(hash));
      map[idx]^=mask;
    }
    inline void unset(word hash){
      word idx=hash/WORDSIZE;
      word mask=((word)1U)<<(MODWORDSIZE(hash));
      if(map[idx] & mask)
	map[idx]&=~mask;
    }
    inline ~bitmap(){
      delete[] map;
    }
  protected:
    word *map;
    word size;
  };
  
  bitmap sorted;

 public:
  static void init(SeqPos _seq_count,word _longestSeq,word _totalHashLength,SeqPos _hash_bits);
  
};

#endif
